@@ -40,7 +40,7 @@ Mail::SpamAssassin - Spam detector and markup engine
 =head1 DESCRIPTION
 
 Mail::SpamAssassin is a module to identify spam using several methods
-including text analysis, internet-based realtime blacklists, statistical
+including text analysis, internet-based realtime blocklists, statistical
 analysis, and internet-based hashing algorithms.
 
 Using its rule base, it uses a wide range of heuristic tests on mail
@@ -64,7 +64,7 @@ use warnings;
 # use bytes;
 use re 'taint';
 
-require 5.006_001;
+require v5.14.0;
 
 use Mail::SpamAssassin::Logger;
 use Mail::SpamAssassin::Constants;
@@ -87,11 +87,10 @@ use Time::HiRes qw(time);
 use Cwd;
 use Config;
 
-our $VERSION = "3.004002";      # update after release (same format as perl $])
+our $VERSION = "4.000001";      # update after release (same format as perl $])
 #our $IS_DEVEL_BUILD = 1;        # 1 for devel build 
 our $IS_DEVEL_BUILD = 0;        # 0 for release versions including rc & pre releases
 
-
 # Used during the prerelease/release-candidate part of the official release
 # process. If you hacked up your SA, you should add a version_tag to your .cf
 # files; this variable should not be modified.
@@ -101,18 +100,18 @@ our @ISA = qw();
 
 # SUB_VERSION is now just <yyyy>-<mm>-<dd>
 our $SUB_VERSION = 'svnunknown';
-if ('$LastChangedDate: 2018-09-13 21:25:10 -0400 (Thu, 13 Sep 2018) $' =~ ':') {
-  # Subversion keyword "$LastChangedDate: 2018-09-13 21:25:10 -0400 (Thu, 13 Sep 2018) $" has been successfully expanded.
+if ('$LastChangedDate: 2024-03-26 17:46:03 +1300 (Tue, 26 Mar 2024) $' =~ ':') {
+  # Subversion keyword "$LastChangedDate: 2024-03-26 17:46:03 +1300 (Tue, 26 Mar 2024) $" has been successfully expanded.
   # Doesn't happen with automated launchpad builds:
   # https://bugs.launchpad.net/launchpad/+bug/780916
-  $SUB_VERSION = (split(/\s+/,'$LastChangedDate: 2018-09-13 21:25:10 -0400 (Thu, 13 Sep 2018) $ updated by SVN'))[1];
+  $SUB_VERSION = (split(/\s+/,'$LastChangedDate: 2024-03-26 17:46:03 +1300 (Tue, 26 Mar 2024) $ updated by SVN'))[1];
 }
 
 
 if (defined $IS_DEVEL_BUILD && $IS_DEVEL_BUILD) {
-  if ('$LastChangedRevision: 1840870 $' =~ ':') {
-    # Subversion keyword "$LastChangedRevision: 1840870 $" has been successfully expanded.
-    push(@EXTRA_VERSION, ('r' . qw{$LastChangedRevision: 1840870 $ updated by SVN}[1]));
+  if ('$LastChangedRevision: 1916544 $' =~ ':') {
+    # Subversion keyword "$LastChangedRevision: 1916544 $" has been successfully expanded.
+    push(@EXTRA_VERSION, ('r' . qw{$LastChangedRevision: 1916544 $ updated by SVN}[1]));
   } else {
     push(@EXTRA_VERSION, ('r' . 'svnunknown'));
   }
@@ -123,7 +122,7 @@ sub Version {
   return join('-', sprintf("%d.%d.%d", $1, $2, $3), @EXTRA_VERSION);
 }
 
-our $HOME_URL = "http://spamassassin.apache.org/";
+our $HOME_URL = "https://spamassassin.apache.org/";
 
 # note that the CWD takes priority.  This is required in case a user
 # is testing a new version of SpamAssassin on a machine with an older
@@ -172,7 +171,7 @@ our @default_userstate_dir = (
 
 ###########################################################################
 
-=item $t = Mail::SpamAssassin->new( { opt => val, ... } )
+=item $t = Mail::SpamAssassin-E<gt>new( { opt =E<gt> val, ... } )
 
 Constructs a new C<Mail::SpamAssassin> object.  You may pass a hash
 reference to the constructor which may contain the following attribute-
@@ -271,7 +270,7 @@ in advance that some information will not be needed by a caller.
 A value of the option can either be a string (a comma-delimited list of tag
 names), or a reference to a list of individual tag names. A caller may provide
 the list in advance, specifying his intention to later collect the information
-through $pms->get_tag() calls. If a name of a tag starts with a 'NO' (case
+through $pms-E<gt>get_tag() calls. If a name of a tag starts with a 'NO' (case
 insensitive), it shows that a caller will not be interested in such tag,
 although there is no guarantee it would save any resources, nor that a tag
 value will be empty. Currently no built-in tags start with 'NO'. A later
@@ -288,9 +287,9 @@ Not requesting it can save a millisecond or two - it mostly serves to
 illustrate the usage of need_tags.
 
 Example:
-  need_tags =>    'TIMING,noLANGUAGES,RELAYCOUNTRY,ASN,noASNCIDR',
+  need_tags =E<gt>    'TIMING,noLANGUAGES,RELAYCOUNTRY,ASN,noASNCIDR',
 or:
-  need_tags => [qw(TIMING noLANGUAGES RELAYCOUNTRY ASN noASNCIDR)],
+  need_tags =E<gt> [qw(TIMING noLANGUAGES RELAYCOUNTRY ASN noASNCIDR)],
 
 =item ignore_site_cf_files
 
@@ -426,9 +425,8 @@ sub new {
     $self->timer_enable();
   }
 
-  $self->{conf} ||= new Mail::SpamAssassin::Conf ($self);
-  $self->{registryboundaries} = Mail::SpamAssassin::RegistryBoundaries->new ($self);
-  $self->{plugins} = Mail::SpamAssassin::PluginHandler->new ($self);
+  $self->{conf} ||= Mail::SpamAssassin::Conf->new($self);
+  $self->{plugins} = Mail::SpamAssassin::PluginHandler->new($self);
 
   $self->{save_pattern_hits} ||= 0;
 
@@ -468,7 +466,7 @@ sub create_locker {
   # for slow but safe, by keeping in quotes
   eval '
     use Mail::SpamAssassin::Locker::'.$class.';
-    $self->{locker} = new Mail::SpamAssassin::Locker::'.$class.' ($self);
+    $self->{locker} = Mail::SpamAssassin::Locker::'.$class.'->new($self);
     1;
   ' or do {
     my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
@@ -546,6 +544,8 @@ sub parse {
   my $msg = Mail::SpamAssassin::Message->new({
     message=>$message, parsenow=>$parsenow,
     normalize=>$self->{conf}->{normalize_charset},
+    body_part_scan_size=>$self->{conf}->{body_part_scan_size},
+    rawbody_part_scan_size=>$self->{conf}->{rawbody_part_scan_size},
     master_deadline=>$master_deadline, suppl_attrib=>$suppl_attrib });
 
   # bug 5069: The goal here is to get rendering plugins to do things
@@ -559,7 +559,7 @@ sub parse {
 
 ###########################################################################
 
-=item $status = $f->check ($mail)
+=item $status = $f-E<gt>check ($mail)
 
 Check a mail, encapsulated in a C<Mail::SpamAssassin::Message> object,
 to determine if it is spam or not.
@@ -584,7 +584,7 @@ sub check {
   $pms;
 }
 
-=item $status = $f->check_message_text ($mailtext)
+=item $status = $f-E<gt>check_message_text ($mailtext)
 
 Check a mail, encapsulated in a plain string C<$mailtext>, to determine if it
 is spam or not.
@@ -610,7 +610,7 @@ sub check_message_text {
 
 ###########################################################################
 
-=item $status = $f->learn ($mail, $id, $isspam, $forget)
+=item $status = $f-E<gt>learn ($mail, $id, $isspam, $forget)
 
 Learn from a mail, encapsulated in a C<Mail::SpamAssassin::Message> object.
 
@@ -661,7 +661,7 @@ sub learn {
 
 ###########################################################################
 
-=item $f->init_learner ( [ { opt => val, ... } ] )
+=item $f-E<gt>init_learner ( [ { opt =E<gt> val, ... } ] )
 
 Initialise learning.  You may pass the following attribute-value pairs to this
 method.
@@ -725,12 +725,15 @@ sub init_learner {
     if (exists $opts->{$k}) { $self->{$v} = $opts->{$k}; }
   }
 
+  # Set flag which can be checked from plugins etc
+  $self->{learning} = 1;
+
   return \%ret;
 }
 
 ###########################################################################
 
-=item $f->rebuild_learner_caches ({ opt => val })
+=item $f-E<gt>rebuild_learner_caches ({ opt =E<gt> val })
 
 Rebuild any cache databases; should be called after the learning process.
 Options include: C<verbose>, which will output diagnostics to C<stdout>
@@ -745,7 +748,7 @@ sub rebuild_learner_caches {
   1;
 }
 
-=item $f->finish_learner ()
+=item $f-E<gt>finish_learner ()
 
 Finish learning.
 
@@ -754,10 +757,11 @@ Finish learning.
 sub finish_learner {
   my $self = shift;
   $self->{bayes_scanner}->force_close(1) if $self->{bayes_scanner};
+  delete $self->{learning};
   1;
 }
 
-=item $f->dump_bayes_db()
+=item $f-E<gt>dump_bayes_db()
 
 Dump the contents of the Bayes DB
 
@@ -768,7 +772,7 @@ sub dump_bayes_db {
   $self->{bayes_scanner}->dump_bayes_db(@opts) if $self->{bayes_scanner};
 }
 
-=item $f->signal_user_changed ( [ { opt => val, ... } ] )
+=item $f-E<gt>signal_user_changed ( [ { opt =E<gt> val, ... } ] )
 
 Signals that the current user has changed (possibly using C<setuid>), meaning
 that SpamAssassin should close any per-user databases it has open, and re-open
@@ -829,7 +833,7 @@ sub signal_user_changed {
   $self->{bayes_scanner}->finish() if $self->{bayes_scanner};
   if ($self->{conf}->{use_bayes}) {
       require Mail::SpamAssassin::Bayes;
-      $self->{bayes_scanner} = new Mail::SpamAssassin::Bayes ($self);
+      $self->{bayes_scanner} = Mail::SpamAssassin::Bayes->new($self);
   } else {
       delete $self->{bayes_scanner} if $self->{bayes_scanner};
   }
@@ -847,13 +851,13 @@ sub signal_user_changed {
 		userstate_dir => $self->{userstate_dir},
 		user_dir => $self->{user_dir},
 	      });
-
+  undef $self->{cf_files_read};
   1;
 }
 
 ###########################################################################
 
-=item $f->report_as_spam ($mail, $options)
+=item $f-E<gt>report_as_spam ($mail, $options)
 
 Report a mail, encapsulated in a C<Mail::SpamAssassin::Message> object, as
 human-verified spam.  This will submit the mail message to live,
@@ -906,7 +910,7 @@ sub report_as_spam {
 
 ###########################################################################
 
-=item $f->revoke_as_spam ($mail, $options)
+=item $f-E<gt>revoke_as_spam ($mail, $options)
 
 Revoke a mail, encapsulated in a C<Mail::SpamAssassin::Message> object, as
 human-verified ham (non-spam).  This will revoke the mail message from live,
@@ -946,73 +950,84 @@ sub revoke_as_spam {
 
 ###########################################################################
 
-=item $f->add_address_to_whitelist ($addr, $cli_p)
+=item $f-E<gt>add_address_to_welcomelist ($addr, $cli_p)
+
+Previously add_address_to_whitelist which will work interchangeably until 4.1.
 
 Given a string containing an email address, add it to the automatic
-whitelist database.
+welcomelist database.
 
 If $cli_p is set then underlying plugin may give visual feedback on additions/failures.
 
 =cut
 
-sub add_address_to_whitelist {
+sub add_address_to_welcomelist {
   my ($self, $addr, $cli_p) = @_;
 
-  $self->call_plugins("whitelist_address", { address => $addr,
+  $self->call_plugins("welcomelist_address", { address => $addr,
                                              cli_p => $cli_p });
 }
+*add_address_to_whitelist = \&add_address_to_welcomelist; # removed in 4.1
 
 ###########################################################################
 
-=item $f->add_all_addresses_to_whitelist ($mail, $cli_p)
+=item $f-E<gt>add_all_addresses_to_welcomelist ($mail, $cli_p)
+
+Previously add_all_addresses_to_whitelist which will work interchangeably until 4.1.
 
 Given a mail message, find as many addresses in the usual headers (To, Cc, From
-etc.), and the message body, and add them to the automatic whitelist database.
+etc.), and the message body, and add them to the automatic welcomelist database.
 
 If $cli_p is set then underlying plugin may give visual feedback on additions/failures.
 
 =cut
 
-sub add_all_addresses_to_whitelist {
+sub add_all_addresses_to_welcomelist {
   my ($self, $mail_obj, $cli_p) = @_;
 
   foreach my $addr ($self->find_all_addrs_in_mail ($mail_obj)) {
-    $self->call_plugins("whitelist_address", { address => $addr,
+    $self->call_plugins("welcomelist_address", { address => $addr,
                                                cli_p => $cli_p });
   }
 }
+*add_all_addresses_to_whitelist = \&add_all_addresses_to_welcomelist; # removed in 4.1
 
 ###########################################################################
 
-=item $f->remove_address_from_whitelist ($addr, $cli_p)
+=item $f-E<gt>remove_address_from_welcomelist ($addr, $cli_p)
+
+Previously remove_address_from_whitelist which will work interchangeably until 4.1.
 
 Given a string containing an email address, remove it from the automatic
-whitelist database.
+welcomelist database.
 
 If $cli_p is set then underlying plugin may give visual feedback on additions/failures.
 
 =cut
 
-sub remove_address_from_whitelist {
+sub remove_address_from_welcomelist {
   my ($self, $addr, $cli_p) = @_;
 
   $self->call_plugins("remove_address", { address => $addr,
                                           cli_p => $cli_p });
 }
+*remove_address_from_whitelist = \&remove_address_from_welcomelist; # removed in 4.1
 
 ###########################################################################
 
-=item $f->remove_all_addresses_from_whitelist ($mail, $cli_p)
+=item $f-E<gt>remove_all_addresses_from_welcomelist ($mail, $cli_p)
+
+Previously remove_all_addresses_from_whitelist which will work interchangeably until 4.1.
 
 Given a mail message, find as many addresses in the usual headers (To, Cc, From
-etc.), and the message body, and remove them from the automatic whitelist
+etc.), and the message body, and remove them from the automatic welcomelist
 database.
 
 If $cli_p is set then underlying plugin may give visual feedback on additions/failures.
 
 =cut
 
-sub remove_all_addresses_from_whitelist {
+sub remove_all_addresses_from_welcomelist {
   my ($self, $mail_obj, $cli_p) = @_;
 
   foreach my $addr ($self->find_all_addrs_in_mail ($mail_obj)) {
@@ -1020,30 +1035,36 @@ sub remove_all_addresses_from_whitelist {
                                             cli_p => $cli_p });
   }
 }
+*remove_all_addresses_from_whitelist = \&remove_all_addresses_from_welcomelist; # removed in 4.1
 
 ###########################################################################
 
-=item $f->add_address_to_blacklist ($addr, $cli_p)
+=item $f-E<gt>add_address_to_blocklist ($addr, $cli_p)
+
+Previously add_address_to_blacklist which will work interchangeably until 4.1.
 
 Given a string containing an email address, add it to the automatic
-whitelist database with a high score, effectively blacklisting them.
+welcomelist database with a high score, effectively blocklisting them.
 
 If $cli_p is set then underlying plugin may give visual feedback on additions/failures.
 
 =cut
 
-sub add_address_to_blacklist {
+sub add_address_to_blocklist {
   my ($self, $addr, $cli_p) = @_;
-  $self->call_plugins("blacklist_address", { address => $addr,
+  $self->call_plugins("blocklist_address", { address => $addr,
                                              cli_p => $cli_p });
 }
+*add_address_to_blacklist = \&add_address_to_blocklist; # removed in 4.1
 
 ###########################################################################
 
-=item $f->add_all_addresses_to_blacklist ($mail, $cli_p)
+=item $f-E<gt>add_all_addresses_to_blocklist ($mail, $cli_p)
+
+Previously add_all_addresses_to_blacklist which will work interchangeably until 4.1.
 
 Given a mail message, find addresses in the From headers and add them to the
-automatic whitelist database with a high score, effectively blacklisting them.
+automatic welcomelist database with a high score, effectively blocklisting them.
 
 Note that To and Cc addresses are not used.
 
@@ -1051,27 +1072,30 @@ If $cli_p is set then underlying plugin may give visual feedback on additions/fa
 
 =cut
 
-sub add_all_addresses_to_blacklist {
+sub add_all_addresses_to_blocklist {
   my ($self, $mail_obj, $cli_p) = @_;
 
   $self->init(1);
 
   my @addrlist;
   my @hdrs = $mail_obj->get_header('From');
-  if ($#hdrs >= 0) {
-    push (@addrlist, $self->find_all_addrs_in_line (join (" ", @hdrs)));
+  foreach my $hdr (@hdrs) {
+    my @addrs = Mail::SpamAssassin::Util::parse_header_addresses($hdr);
+    foreach my $addr (@addrs) {
+      push @addrlist, $addr->{address} if defined $addr->{address};
+    }
   }
 
   foreach my $addr (@addrlist) {
-    $self->call_plugins("blacklist_address", { address => $addr,
+    $self->call_plugins("blocklist_address", { address => $addr,
                                                cli_p => $cli_p });
   }
-
 }
+*add_all_addresses_to_blacklist = \&add_all_addresses_to_blocklist; # removed in 4.1
 
 ###########################################################################
 
-=item $text = $f->remove_spamassassin_markup ($mail)
+=item $text = $f-E<gt>remove_spamassassin_markup ($mail)
 
 Returns the text of the message, with any SpamAssassin-added text (such
 as the report, or X-Spam-Status headers) stripped.
@@ -1155,16 +1179,15 @@ sub remove_spamassassin_markup {
   my $hdrs = $mail_obj->get_pristine_header();
   my $body = $mail_obj->get_pristine_body();
 
-  # remove DOS line endings
-  $hdrs =~ s/\r//gs;
+  # force \n for line-ending processing temporarily
+  $hdrs =~ s/\015?\012/\n/gs;
+  $body =~ s/\015?\012/\n/gs;
 
   # unfold SA added headers, but not X-Spam-Prev headers ...
-  $hdrs = "\n".$hdrs;   # simplifies regexp below
-  1 while $hdrs =~ s/(\nX-Spam-(?!Prev).+?)\n[ \t]+(\S.*\n)/$1 $2/g;
-  $hdrs =~ s/^\n//;
+  1 while $hdrs =~ s/((?:^|\n)X-Spam-(?!Prev).+?)\n[ \t]+(\S.*\n)/$1 $2/g;
 
 ###########################################################################
-  # Backward Compatibilty, pre 3.0.x.
+  # Backward Compatibility, pre 3.0.x.
 
   # deal with rewritten headers w/out X-Spam-Prev- versions ...
   $self->init(1);
@@ -1212,14 +1235,11 @@ sub remove_spamassassin_markup {
   }
 
   # remove any other X-Spam headers we added, will be unfolded
-  $hdrs = "\n".$hdrs;   # simplifies regexp below
-  1 while $hdrs =~ s/\nX-Spam-.*\n/\n/g;
-  $hdrs =~ s/^\n//;
+  1 while $hdrs =~ s/(^|\n)X-Spam-.*\n/$1/g;
 
-  # re-add DOS line endings
-  if ($mail_obj->{line_ending} ne "\n") {
-    $hdrs =~ s/\r?\n/$mail_obj->{line_ending}/gs;
-  }
+  # force original message line endings
+  $hdrs =~ s/\n/$mail_obj->{line_ending}/gs;
+  $body =~ s/\n/$mail_obj->{line_ending}/gs;
 
   # Put the whole thing back together ...
   return join ('', $mbox, $hdrs, $body);
@@ -1227,13 +1247,13 @@ sub remove_spamassassin_markup {
 
 ###########################################################################
 
-=item $f->read_scoreonly_config ($filename)
+=item $f-E<gt>read_scoreonly_config ($filename)
 
 Read a configuration file and parse user preferences from it.
 
 User preferences are as defined in the C<Mail::SpamAssassin::Conf> manual page.
-In other words, they include scoring options, scores, whitelists and
-blacklists, and so on, but do not include rule definitions, privileged
+In other words, they include scoring options, scores, welcomelists and
+blocklists, and so on, but do not include rule definitions, privileged
 settings, etc. unless C<allow_user_rules> is enabled; and they never include
 the administrator settings.
 
@@ -1258,7 +1278,8 @@ sub read_scoreonly_config {
 
   $text = "file start $filename\n" . $text;
   # add an extra \n in case file did not end in one.
-  $text .= "\nfile end $filename\n";
+  $text .= "\n" unless $text =~ /\n\z/;
+  $text .= "file end $filename\n";
 
   $self->{conf}->{main} = $self;
   $self->{conf}->parse_scores_only ($text);
@@ -1269,9 +1290,9 @@ sub read_scoreonly_config {
 
 ###########################################################################
 
-=item $f->load_scoreonly_sql ($username)
+=item $f-E<gt>load_scoreonly_sql ($username)
 
-Read configuration paramaters from SQL database and parse scores from it.  This
+Read configuration parameters from SQL database and parse scores from it.  This
 will only take effect if the perl C<DBI> module is installed, and the
 configuration parameters C<user_scores_dsn>, C<user_scores_sql_username>, and
 C<user_scores_sql_password> are set correctly.
@@ -1295,9 +1316,9 @@ sub load_scoreonly_sql {
 
 ###########################################################################
 
-=item $f->load_scoreonly_ldap ($username)
+=item $f-E<gt>load_scoreonly_ldap ($username)
 
-Read configuration paramaters from an LDAP server and parse scores from it.
+Read configuration parameters from an LDAP server and parse scores from it.
 This will only take effect if the perl C<Net::LDAP> and C<URI> modules are
 installed, and the configuration parameters C<user_scores_dsn>,
 C<user_scores_ldap_username>, and C<user_scores_ldap_password> are set
@@ -1320,10 +1341,10 @@ sub load_scoreonly_ldap {
 
 ###########################################################################
 
-=item $f->set_persistent_address_list_factory ($factoryobj)
+=item $f-E<gt>set_persistent_address_list_factory ($factoryobj)
 
 Set the persistent address list factory, used to create objects for the
-automatic whitelist algorithm's persistent-storage back-end.  See
+automatic welcomelist algorithm's persistent-storage back-end.  See
 C<Mail::SpamAssassin::PersistentAddrList> for the API these factory objects
 must implement, and the API the objects they produce must implement.
 
@@ -1336,7 +1357,7 @@ sub set_persistent_address_list_factory {
 
 ###########################################################################
 
-=item $f->compile_now ($use_user_prefs, $keep_userstate)
+=item $f-E<gt>compile_now ($use_user_prefs, $keep_userstate)
 
 Compile all patterns, load all configuration files, and load all
 possibly-required Perl modules.
@@ -1346,7 +1367,7 @@ plan to fork() or start a new perl interpreter thread to process a message,
 this is suboptimal, as each process/thread will have to perform these actions.
 
 Call this function in the master thread or process to perform the actions
-straightaway, so that the sub-processes will not have to.
+straight away, so that the sub-processes will not have to.
 
 If C<$use_user_prefs> is 0, this will initialise the SpamAssassin
 configuration without reading the per-user configuration file and it will
@@ -1448,7 +1469,7 @@ sub compile_now {
 
 ###########################################################################
 
-=item $f->debug_diagnostics ()
+=item $f-E<gt>debug_diagnostics ()
 
 Output some diagnostic information, useful for debugging SpamAssassin
 problems.
@@ -1468,7 +1489,7 @@ sub debug_diagnostics {
 
 ###########################################################################
 
-=item $failed = $f->lint_rules ()
+=item $failed = $f-E<gt>lint_rules ()
 
 Syntax-check the current set of rules.  Returns the number of 
 syntax errors discovered, or 0 if the configuration is valid.
@@ -1496,7 +1517,7 @@ sub lint_rules {
   $self->{dont_copy_prefs} = $olddcp;       # revert back to previous
 
   # bug 5048: override settings to ensure a faster lint
-  $self->{'conf'}->{'use_auto_whitelist'} = 0;
+  $self->{'conf'}->{'use_auto_welcomelist'} = 0;
   $self->{'conf'}->{'bayes_auto_learn'} = 0;
 
   my $mail = $self->parse(\@testmsg, 1, { master_deadline => undef });
@@ -1513,7 +1534,7 @@ sub lint_rules {
 
 ###########################################################################
 
-=item $f->finish()
+=item $f-E<gt>finish()
 
 Destroy this object, so that it will be garbage-collected once it
 goes out of scope.  The object will no longer be usable after this
@@ -1528,7 +1549,6 @@ sub finish {
   $self->call_plugins("finish_tests", { conf => $self->{conf},
                                         main => $self });
 
-  $self->{conf}->finish(); delete $self->{conf};
   $self->{plugins}->finish(); delete $self->{plugins};
 
   if ($self->{bayes_scanner}) {
@@ -1538,6 +1558,8 @@ sub finish {
 
   $self->{resolver}->finish()  if $self->{resolver};
 
+  $self->{conf}->finish(); delete $self->{conf};
+
   $self->timer_end("finish");
   %{$self} = ();
 }
@@ -1659,6 +1681,11 @@ sub init {
   # Note that this PID has run init()
   $self->{_initted} = $$;
 
+  # if spamd or other forking, wait for spamd_child_init
+  if (!$self->{skip_prng_reseeding}) {
+    $self->set_global_state_dir();
+  }
+
   #fix spamd reading root prefs file
   if (!defined $use_user_pref) {
     $use_user_pref = 1;
@@ -1734,10 +1761,18 @@ sub init {
   }
 
   if ($self->{pre_config_text}) {
-    $self->{config_text} = $self->{pre_config_text} . $self->{config_text};
+    $self->{pre_config_text} .= "\n" unless $self->{pre_config_text} =~ /\n\z/;
+    $self->{config_text} = "file start (pre_config_text)\n".
+                           $self->{pre_config_text}.
+                           "file end (pre_config_text)\n".
+                           $self->{config_text};
   }
   if ($self->{post_config_text}) {
-    $self->{config_text} .= $self->{post_config_text};
+    $self->{post_config_text} .= "\n" unless $self->{post_config_text} =~ /\n\z/;
+    $self->{config_text} .= "\n" unless $self->{config_text} =~ /\n\z/;
+    $self->{config_text} .= "file start (post_config_text)\n".
+                            $self->{post_config_text}.
+                            "file end (post_config_text)\n";
   }
 
   if ($self->{config_text} !~ /\S/) {
@@ -1768,7 +1803,7 @@ sub init {
   # Initialize the Bayes subsystem
   if ($self->{conf}->{use_bayes}) {
       require Mail::SpamAssassin::Bayes;
-      $self->{bayes_scanner} = new Mail::SpamAssassin::Bayes ($self);
+      $self->{bayes_scanner} = Mail::SpamAssassin::Bayes->new($self);
   }
   $self->{'learn_to_journal'} = $self->{conf}->{bayes_learn_to_journal};
 
@@ -1793,6 +1828,21 @@ sub init {
   # should be called only after configuration has been parsed
   $self->{resolver} = Mail::SpamAssassin::DnsResolver->new($self);
 
+  # load GeoDB if some plugin wants it
+  if ($self->{geodb_wanted}) {
+    eval '
+      use Mail::SpamAssassin::GeoDB;
+      $self->{geodb} = Mail::SpamAssassin::GeoDB->new({
+        conf => $self->{conf}->{geodb},
+        wanted => $self->{geodb_wanted},
+      });
+      1;
+    ';
+    if ($@ || !$self->{geodb}) {
+      dbg("config: GeoDB disabled: $@");
+    }
+  }
+
   # TODO -- open DNS cache etc. if necessary
 }
 
@@ -1823,10 +1873,10 @@ sub _read_cf_pre {
       dbg("config: file or directory $path not accessible: $!");
     } elsif (-d _) {
       foreach my $file ($self->$filelistmethod($path)) {
-        $txt .= read_cf_file($file);
+        $txt .= $self->read_cf_file($file);
       }
     } elsif (-f _ && -s _ && -r _) {
-      $txt .= read_cf_file($path);
+      $txt .= $self->read_cf_file($path);
     }
   }
 
@@ -1835,9 +1885,14 @@ sub _read_cf_pre {
 
 
 sub read_cf_file {
-  my($path) = @_;
+  my($self, $path) = @_;
   my $txt = '';
 
+  if ($self->{cf_files_read}->{$path}++) {
+    dbg("config: skipping already read file: $path");
+    return $txt;
+  }
+
   local *IN;
   if (open (IN, "<".$path)) {
 
@@ -1849,7 +1904,8 @@ sub read_cf_file {
 
     $txt = "file start $path\n" . $txt;
     # add an extra \n in case file did not end in one.
-    $txt .= "\nfile end $path\n";
+    $txt .= "\n" unless $txt =~ /\n\z/;
+    $txt .= "file end $path\n";
 
     dbg("config: read file $path");
   }
@@ -1897,7 +1953,7 @@ sub get_and_create_userstate_dir {
     dbg("config: error accessing $fname: $!");
   } else {  # does not exist, create it
     eval {
-      mkpath($fname, 0, 0700);  1;
+      mkpath(Mail::SpamAssassin::Util::untaint_file_path($fname), 0, 0700);  1;
     } or do {
       my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
       dbg("config: mkdir $fname failed: $eval_stat");
@@ -1907,7 +1963,59 @@ sub get_and_create_userstate_dir {
   $fname;
 }
 
-=item $fullpath = $f->find_rule_support_file ($filename)
+# find the most global writable state dir
+# used by dns_block_rule state files etc
+sub set_global_state_dir {
+  my ($self) = @_;
+  # try home_dir_for_helpers
+  my $helper_dir = $self->{home_dir_for_helpers} || '';
+  if ($helper_dir) {
+    my $dir = File::Spec->catdir($helper_dir, ".spamassassin");
+    return if $self->test_global_state_dir($dir);
+  }
+  # try user home (if different from helper home)
+  my $home;
+  if (am_running_on_windows()) {
+    # Windows has a special folder for common appdata (Bug 8050)
+    $home = Mail::SpamAssassin::Util::common_application_data_directory();
+  } else {
+    $home = (Mail::SpamAssassin::Util::portable_getpwuid ($>))[7];
+  }
+  if ($home && $home ne $helper_dir) {
+    my $dir = File::Spec->catdir($home, ".spamassassin");
+    return if $self->test_global_state_dir($dir);
+  }
+  # try LOCAL_STATE_DIR
+  return if $self->test_global_state_dir($self->{LOCAL_STATE_DIR});
+  # fallback to userstate
+  $self->{global_state_dir} = $self->get_and_create_userstate_dir();
+  dbg("config: global_state_dir set to userstate_dir: $self->{global_state_dir}");
+}
+
+sub test_global_state_dir {
+    my ($self, $dir) = @_;
+    eval { mkpath($dir, 0, 0700); }; # just a single stat if exists already
+    # Purge stale test files (enough to do only some times randomly)
+    if (rand() < 0.2 && opendir(WT_DIR, $dir)) {
+      foreach (grep {index($_, '.sawritetest') == 0 &&
+                 (-M File::Spec->catfile($dir, $_)||0) > 0.0001} readdir(WT_DIR)) {
+        unlink(Mail::SpamAssassin::Util::untaint_file_path(File::Spec->catfile($dir, $_)));
+      }
+      closedir WT_DIR;
+    }
+    my $n = ".sawritetest$$".Mail::SpamAssassin::Util::pseudo_random_string(6);
+    my $file = File::Spec->catfile($dir, $n);
+    if (Mail::SpamAssassin::Util::touch_file($file, { create_exclusive => 1 })) {
+      dbg("config: global_state_dir set to $dir");
+      $self->{global_state_dir} = $dir;
+      unlink($file);
+      return 1;
+    }
+    unlink($file); # just in case?
+    return 0;
+}
+
+=item $fullpath = $f-E<gt>find_rule_support_file ($filename)
 
 Find a rule-support file, such as C<languages> or C<triplets.txt>,
 in the system-wide rules directory, and return its full path if
@@ -1920,11 +2028,25 @@ it exists, or undef if it doesn't exist.
 sub find_rule_support_file {
   my ($self, $filename) = @_;
 
+  my @paths;
+  # search custom directories first
+  if ($self->{site_rules_filename}) {
+    foreach my $path (split("\000", $self->{site_rules_filename})) {
+      push @paths, $path  if -d $path;
+    }
+  }
+  if ($self->{rules_filename} && -d $self->{rules_filename}) {
+    push @paths, $self->{rules_filename}
+  }
+  # updates sub-directory missing from @default_rules_path
+  push @paths, '__local_state_dir__/__version__/updates_spamassassin_org';
+  push @paths, @default_rules_path;
+
   return $self->first_existing_path(
-    map { my $p = $_; $p =~ s{$}{/$filename}; $p } @default_rules_path );
+    map { my $p = $_; $p =~ s{$}{/$filename}; $p } @paths );
 }
 
-=item $f->create_default_prefs ($filename, $username [ , $userdir ] )
+=item $f-E<gt>create_default_prefs ($filename, $username [ , $userdir ] )
 
 Copy default preferences file into home directory for later use and
 modification, if it does not already exist and C<dont_copy_prefs> is
@@ -2003,15 +2125,15 @@ sub expand_name {
   if (am_running_on_windows()) {
     my $userprofile = $ENV{USERPROFILE} || '';
 
-    return $userprofile if ($userprofile && $userprofile =~ m/^[a-z]\:[\/\\]/oi);
-    return $userprofile if ($userprofile =~ m/^\\\\/o);
+    return $userprofile if ($userprofile && $userprofile =~ m/^[a-z]\:[\/\\]/i);
+    return $userprofile if ($userprofile =~ m/^\\\\/);
 
-    return $home if ($home && $home =~ m/^[a-z]\:[\/\\]/oi);
-    return $home if ($home =~ m/^\\\\/o);
+    return $home if ($home && $home =~ m/^[a-z]\:[\/\\]/i);
+    return $home if ($home =~ m/^\\\\/);
 
     return '';
   } else {
-    return $home if ($home && $home =~ /\//o);
+    return $home if ($home && index($home, '/') != -1);
     return (getpwnam($name))[7] if ($name ne '');
     return (getpwuid($>))[7];
   }
@@ -2025,6 +2147,9 @@ sub sed_path {
     return $self->{conf}->{sed_path_cache}->{$path};
   }
 
+  # <4.0 compatibility check, to be removed in 4.1
+  my $check_compat = $path eq '__userstate__/auto-welcomelist';
+
   my $orig_path = $path;
 
   $path =~ s/__local_rules_dir__/$self->{LOCAL_RULES_DIR} || ''/ges;
@@ -2032,10 +2157,21 @@ sub sed_path {
   $path =~ s/__def_rules_dir__/$self->{DEF_RULES_DIR} || ''/ges;
   $path =~ s{__prefix__}{$self->{PREFIX} || $Config{prefix} || '/usr'}ges;
   $path =~ s{__userstate__}{$self->get_and_create_userstate_dir() || ''}ges;
+  $path =~ s/__global_state_dir__/$self->{global_state_dir} || ''/ges;
   $path =~ s{__perl_major_ver__}{$self->get_perl_major_version()}ges;
   $path =~ s/__version__/${VERSION}/gs;
   $path =~ s/^\~([^\/]*)/$self->expand_name($1)/es;
 
+  # <4.0 compatibility check, to be removed in 4.1
+  if ($check_compat) {
+    if ($path =~ m{^(.+)/(.+)$}) {
+      # Use auto-whitelist if found
+      if (!-e $path && -e "$1/auto-whitelist") {
+        $path = "$1/auto-whitelist";
+      }
+    }
+  }
+
   $path = Mail::SpamAssassin::Util::untaint_file_path ($path);
   $self->{conf}->{sed_path_cache}->{$orig_path} = $path;
   return $path;
@@ -2074,16 +2210,32 @@ sub get_pre_files_in_dir {
   return $self->_get_cf_pre_files_in_dir($dir, 'pre');
 }
 
+sub _reorder_dir {
+  # Official ASF channel should be loaded first in
+  # order to be able to override scores by using custom channels
+  # bz 7991
+  if($a eq 'updates_spamassassin_org.cf') {
+    return -1;
+  } elsif ($b eq 'updates_spamassassin_org.cf') {
+    return 1;
+  }
+  return $a cmp $b;
+}
+
 sub _get_cf_pre_files_in_dir {
   my ($self, $dir, $type) = @_;
 
   if ($self->{config_tree_recurse}) {
     my @cfs;
-
+    # copied from Mail::SpamAssassin::Util::untaint_file_path
+    # fix bugs 8010 and 8025 by using an untaint pattern that is better on Windows than File::Find's default
+    my $chars = '-_A-Za-z0-9.#%=+,/:()\\@\\xA0-\\xFF\\\\';
+    my $re = qr{^\s*([$chars][${chars}~ ]*)\z};
     # use "eval" to avoid loading File::Find unless this is specified
     eval ' use File::Find qw();
       File::Find::find(
         { untaint => 1,
+          am_running_on_windows() ? (untaint_pattern => $re) : (),
           follow => 1,
           wanted =>
             sub { push(@cfs, $File::Find::name) if /\.\Q$type\E$/i && -f $_ }
@@ -2092,7 +2244,7 @@ sub _get_cf_pre_files_in_dir {
       my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
       die "_get_cf_pre_files_in_dir error: $eval_stat";
     };
-    @cfs = sort { $a cmp $b } @cfs;
+    @cfs = sort { _reorder_dir($a, $b) } @cfs;
     return @cfs;
   }
   else {
@@ -2101,7 +2253,7 @@ sub _get_cf_pre_files_in_dir {
                      /\.${type}$/i && -f "$dir/$_" } readdir(SA_CF_DIR);
     closedir SA_CF_DIR;
 
-    return map { "$dir/$_" } sort { $a cmp $b } @cfs;
+    return map { "$dir/$_" } sort { _reorder_dir($a, $b) } @cfs;
   }
 }
 
@@ -2118,14 +2270,29 @@ sub have_plugin {
 
 sub call_plugins {
   my $self = shift;
+  my $subname = shift;
 
   # We could potentially get called after a finish(), so just return.
   return unless $self->{plugins};
 
+  # Use some calls ourself too
+  if ($subname eq 'spamd_child_init') {
+    # set global dir now if spamd
+    $self->set_global_state_dir();
+  } elsif ($subname eq 'finish_parsing_end') {
+    # Initialize RegistryBoundaries, now that util_rb_tld etc from config is
+    # read.  Plugins can also now use {valid_tlds_re} to one time compile
+    # regexes in finish_parsing_end.
+    $self->{registryboundaries} = Mail::SpamAssassin::RegistryBoundaries->new ($self);
+  } elsif ($subname eq 'whitelist_address' || $subname eq 'blacklist_address') {
+    # Warn about backwards compatibility, removed in 4.1
+    # Third party usage should be rare event, so do not translate function names
+    warn "config: Deprecated $subname called from call_plugins, use welcomelist_address or blocklist_address\n";
+  }
+
   # safety net in case some plugin changes global settings, Bug 6218
   local $/ = $/;  # prevent underlying modules from changing the global $/
 
-  my $subname = shift;
   return $self->{plugins}->callback($subname, @_);
 }
 
@@ -2141,8 +2308,12 @@ sub find_all_addrs_in_mail {
   				Errors-To Mail-Followup-To))
   {
     my @hdrs = $mail_obj->get_header($header);
-    if ($#hdrs < 0) { next; }
-    push (@addrlist, $self->find_all_addrs_in_line(join (" ", @hdrs)));
+    foreach my $hdr (@hdrs) {
+      my @addrs = Mail::SpamAssassin::Util::parse_header_addresses($hdr);
+      foreach my $addr (@addrs) {
+        push @addrlist, $addr->{address} if defined $addr->{address};
+      }
+    }
   }
 
   # find addrs in body, too
@@ -2165,17 +2336,22 @@ sub find_all_addrs_in_mail {
 sub find_all_addrs_in_line {
   my ($self, $line) = @_;
 
+  return () unless defined $line;
+
   # a more permissive pattern based on "dot-atom" as per RFC2822
-  my $ID_PATTERN   = '[-a-z0-9_\+\:\=\!\#\$\%\&\*\^\?\{\}\|\~\/\.]+';
-  my $HOST_PATTERN = '[-a-z0-9_\+\:\/]+';
+  my $ID_PATTERN   = qr/[-a-zA-Z0-9_\+\:\=\!\#\$\%\&\*\^\?\{\}\|\~\/\.]+/;
+  my $HOST_PATTERN = qr/[-a-zA-Z0-9_\+\:\/]+/;
 
   my @addrs;
   my %seen;
   while ($line =~ s/(?:mailto:)?\s*
 	      ($ID_PATTERN \@
-	      $HOST_PATTERN(?:\.$HOST_PATTERN)+)//oix) 
+	      ($HOST_PATTERN(?:\.$HOST_PATTERN)+))//oix) 
   {
     my $addr = $1;
+    my $host = $2;
+    next unless Mail::SpamAssassin::Util::is_fqdn_valid($host);
+    next unless $self->{registryboundaries}->is_domain_valid($host);
     $addr =~ s/^mailto://;
     next if (defined ($seen{$addr})); $seen{$addr} = 1;
     push (@addrs, $addr);
@@ -2196,7 +2372,7 @@ sub sa_die {
 
 ###########################################################################
 
-=item $f->copy_config ( [ $source ], [ $dest ] )
+=item $f-E<gt>copy_config ( [ $source ], [ $dest ] )
 
 Used for daemons to keep a persistent Mail::SpamAssassin object's
 configuration correct if switching between users.  Pass an associative
@@ -2247,7 +2423,7 @@ sub copy_config {
 
 ###########################################################################
 
-=item @plugins = $f->get_loaded_plugins_list ( )
+=item @plugins = $f-E<gt>get_loaded_plugins_list ( )
 
 Return the list of plugins currently loaded by this SpamAssassin object's
 configuration; each entry in the list is an object of type
@@ -2276,8 +2452,8 @@ C<Sys::Syslog>
 
 =head1 MORE DOCUMENTATION
 
-See also E<lt>http://spamassassin.apache.org/E<gt> and
-E<lt>http://wiki.apache.org/spamassassin/E<gt> for more information.
+See also E<lt>https://spamassassin.apache.org/E<gt> and
+E<lt>https://wiki.apache.org/spamassassin/E<gt> for more information.
 
 =head1 SEE ALSO
 
@@ -2288,11 +2464,11 @@ sa-update(1)
 
 =head1 BUGS
 
-See E<lt>http://issues.apache.org/SpamAssassin/E<gt>
+See E<lt>https://issues.apache.org/SpamAssassin/E<gt>
 
 =head1 AUTHORS
 
-The SpamAssassin(tm) Project E<lt>http://spamassassin.apache.org/E<gt>
+The SpamAssassin(tm) Project E<lt>https://spamassassin.apache.org/E<gt>
 
 =head1 COPYRIGHT
 
@@ -2304,6 +2480,6 @@ described in the file C<LICENSE> included with the distribution.
 The latest version of this library is likely to be available from CPAN
 as well as:
 
-  E<lt>http://spamassassin.apache.org/E<gt>
+E<lt>https://spamassassin.apache.org/E<gt>
 
 =cut
